unit main;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  ExtCtrls, ComCtrls, SynMemo, types,
  lexin, relant, lexout, splex,
  DOM, XMLRead, XMLWrite;

type

  { TSPLGRMainForm }

  TSPLGRMainForm = class(TForm)
    ButtonSaveTables: TButton;
    ButtonGenerate: TButton;
    ButtonPISave: TButton;
    ButtonPISaveAs: TButton;
    ButtonPIClose: TButton;
    ButtonPILoad: TButton;
    ButtonPICreate: TButton;
    ButtonScanTest: TButton;
    ButtonTokenAdd: TButton;
    ButtonTokenChange: TButton;
    ButtonTokenDelete: TButton;
    ButtonTokenTest: TButton;
    ComboBoxGenerators: TComboBox;
    EditPIName: TEdit;
    EditTokenID: TLabeledEdit;
    EditTokenRE: TLabeledEdit;
    GeneratorGroup: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    LabelPIFileName: TLabel;
    LexerDefGroup: TGroupBox;
    LexerTestGroup: TGroupBox;
    ListBoxScanToken: TListBox;
    ListBoxTokens: TListBox;
    MemoGenerateOut: TMemo;
    MemoLexerGenOut: TMemo;
    MemoLexerTest: TMemo;
    MemoTokenTest: TMemo;
    OpenDialogProject: TOpenDialog;
    ProjectPC: TPageControl;
    ProjectInfoGroup: TGroupBox;
    SaveDialogProject: TSaveDialog;
    SaveDialogTables: TSaveDialog;
    SynMemoTables: TSynMemo;
    TabIDef: TTabSheet;
    TabGen: TTabSheet;
    TabTest: TTabSheet;
    UpDownToken: TUpDown;
    procedure ButtonGenerateClick(Sender: TObject);
    procedure ButtonPICloseClick(Sender: TObject);
    procedure ButtonPICreateClick(Sender: TObject);
    procedure ButtonPILoadClick(Sender: TObject);
    procedure ButtonPISaveAsClick(Sender: TObject);
    procedure ButtonPISaveClick(Sender: TObject);
    procedure ButtonSaveTablesClick(Sender: TObject);
    procedure ButtonScanTestClick(Sender: TObject);
    procedure ButtonTokenAddClick(Sender: TObject);
    procedure ButtonTokenChangeClick(Sender: TObject);
    procedure ButtonTokenDeleteClick(Sender: TObject);
    procedure ButtonTokenTestClick(Sender: TObject);
    procedure ListBoxTokensDblClick(Sender: TObject);
    procedure TabTestContextPopup(Sender: TObject; MousePos: TPoint;
      var Handled: Boolean);
  private
    { private declarations }
  public
    { public declarations }
    function isID( const s: String ): Boolean;
    procedure ToggleProject( enable: Boolean );

    function TestTokenOpString: Boolean;
    function GetTokenIDFromTokenList( i: Integer ): String;
    function GetTokenREFromTokenList( i: Integer ): String;

    function LoadVerifyProject( const ldf: String ): Boolean;
    function SaveWriteProject( const ldf: String ): Boolean;
  end; 

var
  SPLGRMainForm: TSPLGRMainForm;

implementation

{$R *.lfm}

{ TSPLGRMainForm }

procedure TSPLGRMainForm.ButtonPICreateClick(Sender: TObject);
begin
  if isID( EditPIName.Text ) then
  begin
    ToggleProject( true );
  end else ShowMessage( 'Project Name "'+EditPIName.Text+'" is not a valid FPC Identifier' );
end;

procedure TSPLGRMainForm.ButtonPILoadClick(Sender: TObject);
begin
  if OpenDialogProject.Execute then
  begin
    ToggleProject( true );
    if LoadVerifyProject( OpenDialogProject.FileName ) then
      LabelPIFileName.Caption := OpenDialogProject.FileName
    else
      ToggleProject( false );
  end;
end;

procedure TSPLGRMainForm.ButtonPISaveAsClick(Sender: TObject);
begin
  SaveDialogProject.FileName := EditPIName.Text;
  if SaveDialogProject.Execute then
  begin
    if SaveWriteProject( SaveDialogProject.FileName ) then
    begin
      ShowMessage( 'Saved' );
      LabelPIFileName.Caption := SaveDialogProject.FileName
    end else ShowMessage( 'Saving Project failed' );
  end;
end;

procedure TSPLGRMainForm.ButtonPISaveClick(Sender: TObject);
begin
  if SaveWriteProject( LabelPIFileName.Caption ) then
    ShowMessage( 'Saved' )
  else
    ShowMessage( 'Saving Project failed' );
end;

procedure TSPLGRMainForm.ButtonSaveTablesClick(Sender: TObject);
begin
  case ComboBoxGenerators.ItemIndex of
    0: SaveDialogTables.FileName := EditPIName.Text+'.pas';
    1: SaveDialogTables.FileName := EditPIName.Text+'_nfa.dot';
    2: SaveDialogTables.FileName := EditPIName.Text+'_dfa.dot';
    3: SaveDialogTables.FileName := EditPIName.Text+'_mindfa.dot';
    else SaveDialogTables.FileName := EditPIName.Text;
  end;
  if SaveDialogTables.Execute then
    SynMemoTables.Lines.SaveToFile( SaveDialogTables.FileName );
end;

procedure TSPLGRMainForm.ButtonScanTestClick(Sender: TObject);
var ldef: TASCIILexerDefinition;
    lgen: TSPLexDynamicDFABuilder;
    scanstream: TTextReaderBufferStream;
    scanner: TSPLexScanner;
    i: Integer;
    pmark: Int64;
begin
  MemoLexerGenOut.Clear;
  ListBoxScanToken.Clear;
  try
    ldef := TASCIILexerDefinition.Create( EditPIName.Text );
    try
      for i := 0 to ListBoxTokens.Count-1 do
        ldef.AddToken( GetTokenIDFromTokenList( i ), GetTokenREFromTokenList(i) );
        lgen := TSPLexDynamicDFABuilder.Create( ldef );
        try
          // LabelGenerate table driven lexer
          lgen.VStream := MemoLexerGenOut.Lines;
          lgen.OStream := MemoLexerGenOut.Lines;
          lgen.EStream := MemoLexerGenOut.Lines;
          lgen.Verbose := true;
          lgen.Stats := true;
          lgen.Generate;
          try
            // LabelGenerate stream on memolines
            scanstream := TTextReaderBufferStream.Create( MemoLexerTest.Lines.Text );
            // create a dynamic scanner and scan
            scanner := TSPLexScanner.Create( scanstream, lgen.SPLexDynDFA );
            pmark := -1;
            // scan
            while scanner.NextToken > L_EOS do
            begin
              // position MUST change every step, or well run into an endless loop
              // due to deterministic behavior of the scandfa
              if pmark < scanner.StartPos then
                pmark := scanner.StartPos
              else
                raise Exception.Create( 'Entered Endless Loop. Verify your Tables are Epsilon Free, otherwise File a Bug.' );
              ListBoxScanToken.Items.Add( 'L'+IntToStr( scanner.StartLine ) +
                                          ':C' + IntToStr(scanner.StartCol) +
                                          '.Token ' + lgen.SPLexDynDFA.GetFinalName( scanner.CurrentToken ) );
            end;
            if scanner.CurrentToken <> L_EOS then
            begin
              // no end reached? -> error ^^
              ListBoxScanToken.Items.Add( 'L'+IntToStr( scanner.StartLine ) +
                                          ':C' + IntToStr(scanner.StartCol) +
                                          ', ERROR IN PATTERN' );
            end else
            begin
              // end of stream
              ListBoxScanToken.Items.Add( 'L'+IntToStr( scanner.StartLine ) +
                                          ':C' + IntToStr(scanner.StartCol) +
                                          '.Token ' + lgen.SPLexDynDFA.GetFinalName( scanner.CurrentToken ) );
            end;
          finally
            scanner.ScanTableOwner := false; // lgen frees tables
            scanner.Free;
          end;
        finally
          lgen.Free;
        end;
    finally
      ldef.Free;
    end;
  except on e: Exception do
    begin
      MemoLexerGenOut.Append( 'Exception caught:' );
      MemoLexerGenOut.Append( e.Message );
    end;
  end;
end;

procedure TSPLGRMainForm.ButtonTokenAddClick(Sender: TObject);
begin
  if TestTokenOpString then
    ListBoxTokens.Items.Add( EditTokenID.Text+': '+EditTokenRE.Text );
end;

procedure TSPLGRMainForm.ButtonTokenChangeClick(Sender: TObject);
begin
  if ( ListBoxTokens.ItemIndex >= 0 ) and TestTokenOpString then
    ListBoxTokens.Items[ ListBoxTokens.ItemIndex ] := EditTokenID.Text+': '+EditTokenRE.Text;
end;

procedure TSPLGRMainForm.ButtonTokenDeleteClick(Sender: TObject);
begin
  if ListBoxTokens.ItemIndex >= 0 then
    ListBoxTokens.Items.Delete( ListBoxTokens.ItemIndex );
end;

procedure TSPLGRMainForm.ButtonTokenTestClick(Sender: TObject);
begin
  if TestTokenOpString then
  begin
    MemoTokenTest.Append( 'Test passed' );
  end else
  begin
    MemoTokenTest.Append( 'Test failed' );
  end;
end;

procedure TSPLGRMainForm.ListBoxTokensDblClick(Sender: TObject);
begin
  if ListBoxTokens.ItemIndex >= 0 then
  begin
    EditTokenID.Text := GetTokenIDFromTokenList( ListBoxTokens.ItemIndex );
    EditTokenRE.Text := GetTokenREFromTokenList( ListBoxTokens.ItemIndex );
  end;
end;

procedure TSPLGRMainForm.TabTestContextPopup(Sender: TObject;
  MousePos: TPoint; var Handled: Boolean);
begin

end;

procedure TSPLGRMainForm.ButtonPICloseClick(Sender: TObject);
begin
  ToggleProject( false );
end;

procedure TSPLGRMainForm.ButtonGenerateClick(Sender: TObject);
var ldef: TASCIILexerDefinition;
    lgen: TLexerBuilder;
    i: Integer;
begin
  MemoGenerateOut.Clear;
  SynMemoTables.Clear;
  try
    ldef := TASCIILexerDefinition.Create( EditPIName.Text );
    try
      for i := 0 to ListBoxTokens.Count-1 do
        ldef.AddToken( GetTokenIDFromTokenList( i ), GetTokenREFromTokenList( i ) );
        case ComboBoxGenerators.ItemIndex of
          0: lgen := TSPLexDFABuilder.Create( ldef );
          1:
            begin
              lgen := TNFADotBuilder.Create( ldef );
              TNFADotBuilder(lgen).OutputTypes := [dbtNFA];
            end;
          2:
            begin
              lgen := TNFADotBuilder.Create( ldef );
              TNFADotBuilder(lgen).OutputTypes := [dbtDFA];
            end;
          3:
            begin
              lgen := TNFADotBuilder.Create( ldef );
              TNFADotBuilder(lgen).OutputTypes := [dbtMinDFA];
            end;
          else MemoGenerateOut.Append( 'Unknown Generator: '+IntToStr( ComboBoxGenerators.ItemIndex ) );
        end;
        try
          // LabelGenerate table driven lexer
          lgen.VStream := MemoGenerateOut.Lines;
          lgen.OStream := SynMemoTables.Lines;
          lgen.EStream := MemoGenerateOut.Lines;
          lgen.Verbose := true;
          lgen.Stats := true;
          lgen.Generate;
        finally
          lgen.Free;
        end;
    finally
      ldef.Free;
    end;
  except on e: Exception do
    begin
      MemoGenerateOut.Append( 'Exception caught:' );
      MemoGenerateOut.Append( e.Message );
    end;
  end;
end;

function TSPLGRMainForm.isID(const s: String): Boolean;
var i: Integer;
begin
  Result := true;
  if Length( s ) > 0 then
  begin
    if not ( s[ 1 ] in ['A'..'Z','a'..'z'] ) then Exit(false);
    if Length( s ) > 1 then
    begin
      i := 2;
      while i <= Length( s ) do
      begin
        if not ( s[ i ] in ['A'..'Z','a'..'z','0'..'9','_'] ) then Exit(false);
        Inc( i, 1 );
      end;
    end;
  end else Exit(false);
end;

procedure TSPLGRMainForm.ToggleProject(enable: Boolean);
begin
  // toggle enabled
  EditPIName.Enabled := not enable;
  ButtonPICreate.Enabled := not enable;
  ButtonPILoad.Enabled := not enable;
  ButtonPISave.Enabled := enable;
  ButtonPISaveAs.Enabled := enable;
  ButtonPIClose.Enabled := enable;
  LexerDefGroup.Enabled := enable;
  LexerTestGroup.Enabled := enable;
  GeneratorGroup.Enabled := enable;
  // clear all lists
  MemoTokenTest.Clear;
  MemoGenerateOut.Clear;
  MemoLexerTest.Clear;
  MemoLexerGenOut.Clear;
  ListBoxScanToken.Clear;
  ListBoxTokens.Clear;
  SynMemoTables.Clear;
end;

function TSPLGRMainForm.TestTokenOpString: Boolean;
var ldef: TASCIILexerDefinition;
begin
  MemoTokenTest.Clear;
  Result := isID( EditTokenID.Text );
  if Result then
  begin
    Result := false;
    try
      ldef := TASCIILexerDefinition.Create( EditPIName.Text );
      try
        ldef.AddToken( EditTokenID.Text, EditTokenRE.Text );
        MemoTokenTest.Append( 'Token Parsed as: ' + RegExToOpString( ldef.Tokens.RegExs[ 0 ] ) );
        Result := true;
      finally
        ldef.Free;
      end;
    except on e: Exception do
      begin
        MemoTokenTest.Append( 'Conversion failed: '+e.Message );
      end;
    end;
  end else
    MemoTokenTest.Append( 'Token Identifier is not a valid FPC Identifier' );
end;

function TSPLGRMainForm.GetTokenIDFromTokenList(i: Integer): String;
var e: Integer;
begin
  // e := Pos( ListBoxTokens.Items[i], ':' );
  e := 1;
  while e <= Length( ListBoxTokens.Items[ i ] ) do
  begin
    if ListBoxTokens.Items[ i ][ e ] = ':' then Break;
    Inc( e, 1 );
  end;
  Result := Copy( ListBoxTokens.Items[ i ], 1, e-1 );
end;

function TSPLGRMainForm.GetTokenREFromTokenList(i: Integer): String;
var s: Integer;
begin
  //s := Pos( ListBoxTokens.Items[i], ':' );
  s := 1;
  while s <= Length( ListBoxTokens.Items[ i ] ) do
  begin
    if ListBoxTokens.Items[ i ][ s ] = ':' then Break;
    Inc( s, 1 );
  end;
  Result := Copy( ListBoxTokens.Items[ i ], s+2, Length( ListBoxTokens.Items[ i ] )-(s+1) );
end;

function TSPLGRMainForm.LoadVerifyProject(const ldf: String): Boolean;
var ldfdoc: TXMLDocument;
    ldfroot, ldfnode: TDOMNode;
begin
  ldfdoc := nil;
  Result := true;
  try
    ReadXMLFile( ldfdoc, ldf );
    try
      ldfroot := ldfdoc.FindNode( 'ldf' );
      if not Assigned( ldfroot ) then
        raise Exception.Create('Not a valid Lexer Description File');

      ldfnode := ldfroot.FindNode( 'Project' );
      if Assigned( ldfnode ) then
      begin
        EditPIName.Text := ldfnode.TextContent;
        if not isID( EditPIName.Text ) then
          raise Exception.Create('Corrupted or Wrong file');
      end else raise Exception.Create('Not a valid Lexer Description File');

      ldfnode := ldfroot.FindNode( 'TokenList' );
      if Assigned( ldfnode ) then
      begin
        ListBoxTokens.Items.DelimitedText := ldfnode.TextContent;
        //TODO: validate items
      end else raise Exception.Create('Not a valid Lexer Description File');

      ldfnode := ldfroot.FindNode( 'TokenTest' );
      if Assigned( ldfnode ) then
      begin
        MemoLexerTest.Lines.DelimitedText := ldfnode.TextContent;
      end else raise Exception.Create('Not a valid Lexer Description File');
    finally
      ldfdoc.Free;
    end;
  except on e: Exception do
    begin
      ShowMessage( 'Loading failed: '+e.Message );
      Result := false;
    end;
  end;
end;

function TSPLGRMainForm.SaveWriteProject(const ldf: String): Boolean;
var ldfdoc: TXMLDocument;
    ldfroot, ldfnode: TDOMElement;
begin
  Result := true;
  ldfdoc := TXMLDocument.Create;
  try
    try
      ldfroot := ldfdoc.CreateElement( 'ldf' );
      ldfdoc.AppendChild( ldfroot );

      ldfnode := ldfdoc.CreateElement( 'Project' );
      ldfnode.TextContent := EditPIName.Text;
      ldfroot.AppendChild( ldfnode );

      ldfnode := ldfdoc.CreateElement( 'TokenList' );
      ldfnode.TextContent := ListBoxTokens.Items.DelimitedText;
      ldfroot.AppendChild( ldfnode );

      ldfnode := ldfdoc.CreateElement( 'TokenTest' );
      ldfnode.TextContent := MemoLexerTest.Lines.DelimitedText;
      ldfroot.AppendChild( ldfnode );

      WriteXMLFile( ldfdoc, ldf );
    finally
      ldfdoc.Free;
    end;
  except on e: Exception do
    Result := false;
  end;
end;

end.

