unit lexin;

{$mode delphi}{$H+}

interface

uses sysutils, relant;

type
  TRegExs = array of TRegExString;

  { TRegExList }

  TRegExList = class
    private
      FRegExs: TRegExs;
    public
      property RegExs: TRegExs read FRegExs;
      constructor Create;
      destructor Destroy; override;
      procedure AddRegEx( ARegEx: TRegExString );
  end;

  TInputSet = class
    public
      function ToString: String; virtual; abstract;
      function ToNFA: TUNFA; virtual; abstract;
  end;

  { TLexerDefinition }

  TLexerDefinition = class
    private
      FTokens: TRegExList;
      FLexerName: String;
    protected
      procedure AddToken( ARegEx: TRegExString );
    public
      property Tokens: TRegExList read FTokens;
      property LexerName: String read FLexerName ;
      constructor Create( const ALexerName: String ); virtual;
      destructor Destroy; override;
  end;

 (* ASCII (8bit codepage dependend char) Input*)

  TASCIIAlphaSet = set of Char;

  TASCIIInputSet = class( TInputSet )
    private
      FSet: TASCIIAlphaSet;
    public
      function SetToString( stripapo: Boolean ): String;
      function AlphaSetToString: String;
      function AlphaSetToStringStripApo: String;

      constructor Create( const AAlphaSet: TASCIIAlphaSet ); overload;
      constructor Create( AAlpha: Char ); overload;
      function ToString: String; override;
      function ToNFA: TUNFA; override;
  end;

  TASCIILexerDefinition = class( TLexerDefinition )
    private
      function parseCharCode( var p: Integer;
        const TokenRegEx: String; RegEx: TRegExString ): Char;
      function parseSet( var p: Integer;
        const TokenRegEx: String; RegEx: TRegExString ): TASCIIAlphaSet;
      function parseRegEx( const tid, TokenRegEx: String ): TRegExString;
    public
      procedure AddToken( const TokenIdent, TokenRegEx: String ); overload;
  end;

implementation

{ TRegExList }

constructor TRegExList.Create;
begin
  SetLength( FRegExs, 0 );
end;

destructor TRegExList.Destroy;
var i: Integer;
begin
  for i := Low( FRegExs ) to High( FRegExs ) do
   FRegExs[ i ].Free;
  inherited Destroy;
end;

procedure TRegExList.AddRegEx(ARegEx: TRegExString);
begin
  SetLength( FRegExs, Succ( Length( FRegExs ) ) );
  FRegExs[ High( FRegExs ) ] := ARegEx;
end;

{ TLexerDefinition }

procedure TLexerDefinition.AddToken(ARegEx: TRegExString);
begin
  FTokens.AddRegEx( ARegEx );
end;

constructor TLexerDefinition.Create(const ALexerName: String);
begin
  FLexerName := ALexerName;
  FTokens := TRegExList.Create;
end;

destructor TLexerDefinition.Destroy;
begin
  FTokens.Free;
  inherited Destroy;
end;

function TASCIIInputSet.SetToString( stripapo: Boolean ): String;
var il: Boolean;
    ls, l, c: char;
    tmps: String;

  function apogen( ngen: Boolean ): String;
  begin
    if ngen then Result := '' else Result := '''';
  end;

  function sqtd( c: Char ): String;
  begin
    if c = apogen( false ) then
      result := apogen( false ) + apogen( stripapo )
    else
      result := c;
  end;

begin
  Result := '';
  il := false;
  c := #0;
  l := #0;
  ls := #0;
  while true do
  begin
    if c in FSet then
    begin
      ls := c;
      while ( c in FSet ) do
      begin
        l := c;
        if c = #255 then break;
        Inc( c, 1 );
      end;
      il := true;
    end;
    if il then
    begin
      if Length( Result ) > 0 then Result := Result + ',';
      if ls = l then
      begin
        if ( l < #33 ) or ( l > #126 ) then
        begin
          Str( Ord( l ), tmps );
          Result := Result + '#' + tmps;
        end else
        begin
          Result := Result + apogen( stripapo ) + sqtd( l ) + apogen( stripapo );
        end;
      end else
      begin
        if ( ls < #33 ) or ( ls > #126 ) then
        begin
          Str( Ord( ls ), tmps );
          Result := Result + '#' + tmps;
        end else
        begin
          Result := Result + apogen( stripapo ) + sqtd( ls ) + apogen( stripapo );
        end;
        if ( l < #33 ) or ( l > #126 ) then
        begin
          Str( Ord( l ), tmps );
          Result := Result + '..' + '#' + tmps;
        end else
        begin
          Result := Result + '..' + apogen( stripapo ) + sqtd( l ) + apogen( stripapo );
        end;
      end;
      il := false;
    end;
    if c = #255 then break;
    Inc( c, 1 );
  end;
end;

function TASCIIInputSet.AlphaSetToString: String;
begin
  Result := SetToString( false );
end;

function TASCIIInputSet.AlphaSetToStringStripApo: String;
begin
  Result := SetToString( true );
end;

constructor TASCIIInputSet.Create( const AAlphaSet: TASCIIAlphaSet );
begin
  FSet := AAlphaSet;
end;

constructor TASCIIInputSet.Create( AAlpha: Char );
begin
  FSet := [ AAlpha ];
end;

function TASCIIInputSet.ToString: String;
begin
  Result := '['+AlphaSetToString+']';
end;

function TASCIIInputSet.ToNFA: TUNFA;
begin
  Result := TUNFA.Create(  ToString, TAlphaSet( FSet ) );
end;

{ TASCIILexerDefinition }

function TASCIILexerDefinition.parseCharCode( var p: Integer;
  const TokenRegEx: String; RegEx: TRegExString ): Char;

var tstr: String;
    ti: Integer;
    a: Integer;
    e: Word;
    CCharSet: TASCIIAlphaSet;

begin
  // parse val convertable string
  Inc( p, 1 );
  if ( Length( TokenRegEx ) >= succ( p ) ) and
     ( TokenRegEx[ p ] in [ '$','&','%' ] ) then
  begin
    case TokenRegEx[ p ] of
      '$': begin
             ti := p + 2; // Hex Code
             CCharSet := [ '0' .. '9', 'a' .. 'f', 'A'..'F' ];
           end;
      '&': begin
             ti := p + 3; // Oct Code
             CCharSet := [ '0' .. '7' ];
           end;
      '%': begin
             ti := p + 8; // Bin Code
             CCharSet := [ '0', '1' ];
           end;
    end;
    tstr := TokenRegEx[ p ];
    Inc( p, 1 );
  end else
  begin
    ti := p + 2; // Dec Code
    tstr := '';
    CCharSet := [ '0'..'9' ];
  end;
  while ( p <= ti ) and
        ( p <= Length( TokenRegEx ) ) and
        ( TokenRegEx[ p ] in CCharSet ) do
  begin
    tstr := tstr + TokenRegEx[ p ];
    Inc( p, 1 );
  end;
  Dec( p, 1 );
  // transform
  a := 0;
  Val( tstr, a, e );
  if ( e <> 0 ) or ( a < 0 ) or ( a > Ord( High( Char ) ) ) then
  begin
    RegEx.Free;
    raise Exception.Create( 'Character Code Parsing failed' );
  end;
  Result := Chr( a );
end;

// numcode processing ist faulty

function TASCIILexerDefinition.parseSet( var p: Integer;
  const TokenRegEx: String; RegEx: TRegExString ): TASCIIAlphaSet;
var invert, escape, second : Boolean;
    tc, tc2: Char;
    resset: TASCIIAlphaSet;
begin
  resset := [];
  Inc( p, 1 );
  escape := false;
  if p <= Length( TokenRegEx ) then
  begin
    if TokenRegEx[ p ] <> ']' then  // [] -> empty
    begin
      invert := ( TokenRegEx[ p ] = '^'); // [^...] inverted set
      if invert then Inc( p, 1 );
      second := ( p <= Length( TokenRegEx ) ) and ( TokenRegEx[ p ] <> ']' );
      while second do
      begin
        tc := TokenRegEx[ p ];
        if not escape then
        begin
          if tc <> '\' then
          begin
            if tc = '#' then tc := ParseCharCode( p, TokenRegEx, RegEx );
            if ( ( ( p <  pred( Length( TokenRegex ) ) ) and
                   ( TokenRegex[ succ( succ( p ) ) ] <> ']' ) ) or
                 ( ( p <  pred( pred( Length( TokenRegex ) ) ) ) and
                   ( TokenRegex[ succ( succ( p ) ) ] = '\' ) ) ) and
               ( TokenRegex[ succ( p ) ] = '-' ) then
            begin
              Inc( p, 2 );
              tc2 := TokenRegex[ p ];
              if tc2 <> '\' then
              begin
                if tc2 = '#' then tc2 := ParseCharCode( p, TokenRegEx, RegEx );
              end else
              begin
                Inc( p, 1);
                tc2 := TokenRegex[ p ];
              end;
              if ord( tc ) <= ord( tc2 ) then
              begin
                for tc := tc to tc2 do resset := resset + [ tc ];
              end else
              begin
                for tc := tc downto tc2 do resset := resset + [ tc ];
              end;
            end else
            begin
              resset := resset + [ tc ];
              Inc( p, 1 );
            end;
          end else
          begin
            escape := true;
            Inc( p, 1 );
          end;
        end else
        begin
          resset := resset + [ tc ];
          escape := false;
          Inc( p, 1 );
        end;
        if not escape then
        begin
          second := ( p <= Length( TokenRegEx ) ) and ( TokenRegEx[ p ] <> ']' );
        end else
        begin
          second := ( p <= Length( TokenRegEx ) );
        end;
      end;
      if invert then resset := [ #0..#255 ] - resset;
      if ( Length( TokenRegEx ) < p ) then
      begin
        RegEx.Free;
        raise Exception.Create( 'Set Parsing failed, missing end of Set' );
      end;
    end;
  end else
  begin
    RegEx.Free;
    raise Exception.Create( 'Set Parsing failed, missing end of Set' );
  end;
  Result := resset;
end;

function TASCIILexerDefinition.parseRegEx( const tid, TokenRegEx: String ): TRegExString;
var p, parl: Integer;
    aset: TASCIIInputSet;
begin
  Result := TRegExString.Create( tid );
  p := 1;
  parl := 0;
  try
    while p <= Length( TokenRegEx ) do
    begin
      case TokenRegEx[ p ] of
        '(':
          begin
            Inc(parl,1);
            Result.AddOperation( reoPatStart );
          end;
        ')':
          begin
            Dec( parl, 1 );
            if parl < 0 then raise Exception.Create('No Matching ( for )');
            Result.AddOperation( reoPatEnd );
          end;
        '*': Result.AddOperation( reoRepeat );
        '+': Result.AddOperation( reoMinOne );
        '?': Result.AddOperation( reoMaxOne );
        '|': Result.AddOperation( reoOr );
        '\': begin
               Inc( p, 1 );
               if p <= Length( TokenRegEx ) then
               begin
                 aset := TASCIIInputSet.Create( TokenRegEx[ p ] );
                 Result.AddUNFA( aset.ToNFA );
                 aset.Free;
               end else raise Exception.Create( 'Nothing to Escape' );
             end;
        '[': begin
               aset := TASCIIInputSet.Create(
                 parseSet( p, TokenRegEx, Result ) );
               Result.AddUNFA( aset.ToNFA );
               aset.Free;
             end;
        '#': begin
               aset := TASCIIInputSet.Create(
                 parseCharCode( p, TokenRegEx, Result ) );
               Result.AddUNFA( aset.ToNFA );
               aset.Free;
             end;
        else
          begin
            aset := TASCIIInputSet.Create( TokenRegEx[ p ] );
            Result.AddUNFA( aset.ToNFA );
            aset.Free;
          end;
      end;
      Inc( p, 1 );
    end;
    if parl <> 0 then raise Exception.Create('No Matching ) for (');
  except on Exception do
    begin
      Result.Free;
      raise;
    end;
  end;
end;

procedure TASCIILexerDefinition.AddToken( const TokenIdent, TokenRegEx: String );
begin
  AddToken( parseRegEx( TokenIdent, TokenRegEx ) );
end;


end.
